//
//    File: mview.cc
//
//    (C) 2000 Helmut Cantzler
//
//    Licensed under the terms of the Lesser General Public License.
//

#include <qspinbox.h> 
#include <qslider.h>
#include <qtoolbar.h>
#include <qtoolbutton.h>
#include <qpopupmenu.h>
#include <qmenubar.h>
#include <qkeycode.h>
#include <qfile.h>
#include <qfiledialog.h>
#include <qprogressdialog.h>
#include <qstatusbar.h>
#include <qmessagebox.h>
#include <qapplication.h>
#include <qaccel.h>
#include <qpixmap.h>
#include <qimage.h>
#include <qhbox.h>
#include <qtooltip.h>
#include <qcolordialog.h>
#include <qdesktopwidget.h>

#include "mview.h"
#include "viewpointdialog.h"
#include "coordinates.h"

#ifdef TEXT_MAP
  #include "texture_image.h"
#endif

#include "p_mesh.h"
#include "gts_mesh.h"
#include "geomview_mesh.h"
#include "vrml1_mesh.h"
#include "vrml2_mesh.h"
#include "list_mesh.h"
#include "feature_mesh.h"
#include "ply_mesh.h"

#include "icons/icon_fileopen.xpm"
#include "icons/icon_texture.xpm"
#include "icons/icon_solid.xpm"
#include "icons/icon_wire.xpm"
#include "icons/icon_point.xpm"
#include "icons/icon_front.xpm"
#include "icons/icon_feature.xpm"
#include "icons/icon_coordinates.xpm"

#ifdef TEXT_MAP
  #include "icons/icon_mapping.xpm"
#endif

QProgressDialog *pd;
QLabel *label_mesh, *label_features, *label_fps;

int update_progress(int pos)
{
  static int q=0;

  if (q % 100 == 0)
    pd->setProgress(pos);
  q++;

  return pd->wasCancelled();
}

void set_total(int size)
{
  pd->setTotalSteps(size);
}

MView::MView(int argc, char **argv)
    : QMainWindow( 0, "Mesh Viewer")
{
  //// Set OpenGL options

  QGLFormat f;
  f.setDoubleBuffer(TRUE);
  f.setDepth(TRUE);
  f.setStencil(TRUE);
  QGLFormat::setDefaultFormat(f);

  //// Create the main and the mesh widget

  QHBox *q = new QHBox(this);
  q->setMargin(10);
  q->setSpacing(10);
  
  // set defaults
  set_glmesh_defaults();

  // read files
  if (argc >= 2)
    {
      name = argv[1];
      load_mesh_file(argv[1], &settings->mesh);

      if (settings->mesh != NULL)
	{
	  // Map vertices and triangle centroids to mesh centroid
	  settings->mesh->move_to_centre();
	  settings->mesh->scale_into_normal_sphere();
	}
    }      

  if (argc >= 3)
    load_mesh_file(argv[2], &settings->features);

  glmesh = new GLMesh(q, "GL Mesh", *settings);

  /*
  printf("depth: %d\n", glmesh->format().depth() ? 1 : 0);
  printf("rgba: %d\n", glmesh->format().rgba() ? 1 : 0);
  printf("alpha: %d\n", glmesh->format().alpha() ? 1 : 0);
  printf("accum: %d\n", glmesh->format().accum() ? 1 : 0);
  printf("stencil: %d\n", glmesh->format().stencil() ? 1 : 0);
  printf("stereo: %d\n", glmesh->format().stereo() ? 1 : 0);
  printf("directRendering: %d\n", glmesh->format().directRendering() ? 1 : 0);
  printf("hasOverlay: %d\n", glmesh->format().hasOverlay() ? 1 : 0);
  */

  //// Set the main layout ////

  // from, until, step size, start value, orientation, parent, text
  clipping = new QSlider(-100, 100, 10, 0, Vertical,
				  q, "clipping slider");
  QToolTip::add(clipping, "Clipping");
  connect(clipping, SIGNAL(valueChanged(int)), glmesh, SLOT(setClipping(int)));

  setCentralWidget(q);
  
  //// Set the tool bar

  QToolBar *tools = new QToolBar(this, "MView Toolbar");

  new QToolButton(QPixmap(icon_fileopen), "Open Mesh", QString::null,
		  this, SLOT(load_mesh()), tools, "open file");

  tools->addSeparator();

  new QToolButton(QPixmap(icon_texture), "Texture", QString::null,
		  this, SLOT(setTexture()), tools, "texture");
  new QToolButton(QPixmap(icon_solid), "Solid", QString::null,
		  this, SLOT(setSolid()), tools, "solid");
  new QToolButton(QPixmap(icon_front), "Frontlines", QString::null,
		  this, SLOT(setFrontlines()), tools, "frontlines");
  new QToolButton(QPixmap(icon_wire), "Wireframe", QString::null,
		  this, SLOT(setWire()), tools, "wireframe");
  new QToolButton(QPixmap(icon_point), "Points", QString::null,
		  this, SLOT(setPoints()), tools, "points");
  new QToolButton(QPixmap(icon_feature), "Features", QString::null,
  		  this, SLOT(setFeatures()), tools, "features");

  tools->addSeparator();

  pick = new QToolButton(QPixmap(icon_coordinates), "Pick", QString::null,
  			 this, SLOT(togglePick()), tools, "pick");
  pick->setToggleButton(TRUE);

#ifdef TEXT_MAP
  text = new QToolButton(QPixmap(icon_mapping), "Mapping", QString::null,
  			 this, SLOT(toggleTextureMapping()), tools, "mapping");
  text->setToggleButton(TRUE);
#endif

  tools->addSeparator();

  QSpinBox *lighting = new QSpinBox(1, 20, 1, tools, "lighting spinbox");
  QToolTip::add(lighting, "Light intensity");
  lighting->setValue(10);
  connect(lighting, SIGNAL(valueChanged(int)),
	  glmesh, SLOT(setLightIntensity(int)));

  QSpinBox *line_width = new QSpinBox(1, 20, 1, tools, "line width spinbox");
  QToolTip::add(line_width, "Line width");
  line_width->setValue(1);
  connect(line_width, SIGNAL(valueChanged(int)),
	  glmesh, SLOT(setLineSize(int)));

  QSpinBox *point_size = new QSpinBox(1, 20, 1, tools, "point size spinbox");
  QToolTip::add(point_size, "Point size");
  point_size->setValue(1);
  connect(point_size, SIGNAL(valueChanged(int)),
	  glmesh, SLOT(setPointSize(int)));

  //// Set the menu ////
  
  QStrList format = QImage::outputFormats();
  screenshot = new QPopupMenu(this);
  for (const char* f=format.first(); f; f=format.next())
    screenshot->insertItem(f);
  connect(screenshot, SIGNAL(activated(int)), this, SLOT(takeShot(int)) );

  save = new QPopupMenu(this);
  save->insertItem("P Mesh");
  save->insertItem("GTS");
  save->insertItem("Geomview");
  save->insertItem("VRML 1");
  save->insertItem("VRML 2");
	save->insertItem("PLY");
  connect(save, SIGNAL(activated(int)), this, SLOT(save_as(int)) );

  QPopupMenu *file = new QPopupMenu(this);
  menuBar()->insertItem("&File", file);
  file->insertItem("&Open Mesh", this, SLOT(load_mesh()), Key_O);
  file->insertItem("Open &Features", this, SLOT(load_feature()));
  file->insertSeparator();
  file->insertItem("&Save as", save);
  file->insertSeparator();
  file->insertItem("&Take Screenshot", screenshot);
  file->insertSeparator();
  file->insertItem("&Quit", qApp, SLOT( quit() ), Key_Escape);
  
  QPopupMenu *edit = new QPopupMenu(this);
  menuBar()->insertItem("&Edit", edit);
  edit->insertItem("&Negate normals", this, SLOT(negateNormals()));
  edit->insertItem("&Remove double points", this, SLOT(removeDoublePoints()));
  
  ids= new int[SWITCHES];
  
  modes = new QPopupMenu(this);
  menuBar()->insertItem("&Display", modes);
  modes->setCheckable(TRUE);
  ids[MVIEW_TEXTURE]=modes->insertItem("&Texture", this,
			  SLOT(setTexture()), Key_1);
  ids[MVIEW_SOLID]=modes->insertItem("S&olid", this,
			       SLOT(setSolid()), Key_2);
  ids[MVIEW_FRONTLINES]=modes->insertItem("F&rontlines", this,
				    SLOT(setFrontlines()), Key_3);
  ids[MVIEW_WIRE]=modes->insertItem("&Wireframe", this,
			      SLOT(setWire()), Key_4);
  ids[MVIEW_POINTS]=modes->insertItem("Po&ints", this,
				SLOT(setPoints()), Key_5);
  ids[MVIEW_FEATURES]=modes->insertItem("F&eatures", this,
				  SLOT(setFeatures()), Key_6);
  modes->insertSeparator();
  ids[MVIEW_PICK]=modes->insertItem("&Pick object", this,
			      SLOT(togglePick()), Key_P);
#ifdef TEXT_MAP
  ids[MVIEW_TEXT]=modes->insertItem("&Texture image", this,
			       SLOT(toggleTextureMapping()));
#endif
  modes->insertSeparator();
  ids[MVIEW_LIGHTING]=modes->insertItem("&Lighting", this,
				  SLOT(toggleLighting()), Key_L);
  ids[MVIEW_SHAPE]=modes->insertItem("&Shape colors", this,
			       SLOT(toggleShapeColors()), Key_S);
  ids[MVIEW_NORMALS]=modes->insertItem("Surface &normals", this,
				 SLOT(toggleSurfaceNormals()), Key_N);
  ids[MVIEW_CUT]=modes->insertItem("&Cut back faces", this,
			     SLOT(toggleCutBackFaces()), Key_C);
  ids[MVIEW_FILTER]=modes->insertItem("&Bilinear filtering", this,
				SLOT(toggleBilinearFiltering()), Key_B);
  ids[MVIEW_FEATURE]=modes->insertItem("&Features", this,
				 SLOT(toggleFeatures()), Key_F);
  ids[MVIEW_RATIO]=modes->insertItem("&Keep aspect ratio", this,
			       SLOT(toggleAspectRatio()), Key_K);
  modes->insertSeparator();
  modes->insertItem("Back&ground color...", this,
		    SLOT(setBackgroundColor()), Key_G);
  
  viewpoint = new QPopupMenu(this);
  menuBar()->insertItem("&Viewpoint", viewpoint);
  viewpoint->insertItem("&Set", this, SLOT(set_view()) );
  viewpoint->insertItem("&Reset", this, SLOT(reset_view()) );
  viewpoint->insertSeparator();
  viewpoint->insertItem("S&ave", this, SLOT(save_view()) );
  viewpoint->insertItem("&Clear", this, SLOT(clear_views()) );
  viewpoint->insertSeparator();
  
  QPopupMenu *help = new QPopupMenu(this);
  menuBar()->insertItem("&Help", help);
  help->insertItem("&About", this, SLOT(about()));
  help->insertItem("About &Qt", this, SLOT(aboutQt()));

  //// Set status bar ////

  label_mesh = new QLabel(statusBar(), "Mesh Label");
  statusBar()->addWidget(label_mesh, 1, TRUE);
  label_features = new QLabel(statusBar(), "Feature Label");
  statusBar()->addWidget(label_features, 1, TRUE);
  label_fps = new QLabel(" FPS: 0.0 ", statusBar(), "FPS Label");
  statusBar()->addWidget(label_fps, 0, TRUE);

  // for updating label_fps
  startTimer(3000); // 3 seconds

  //// Ready ////

  setCaption("Mesh Viewer");
  resize(580, 620);
  set_mview_defaults();
  set_status();
}

MView::~MView()
{
  delete settings->mesh;
  delete settings->features;
  delete settings;
  delete ids;
}

// Slots

void MView::reset_view()
{
  GLMesh_Settings s = glmesh->getSettings();

  s.xShift = s.yShift = s.zShift = 0.0;
  s.clipping = 0.0;

  // create identity matrix
  for (int i=0; i < 4; i++)
    for (int j=0; j < 4; j++)
      s.tb_transform[i][j]= i == j ? 1.0 : 0.0;

  glmesh->setViewpoint(s);
}

void MView::set_view()
{
  GLMesh_Settings s = glmesh->getSettings();
  ViewpointDialog vpd(this, &s, "Viewpoint");

  if (vpd.exec() == QDialog::Accepted)
    glmesh->setViewpoint(s);
}

void MView::clear_views()
{
  QMap<int, GLMesh_Settings>::Iterator it;

  for(it = viewpoints.begin(); it != viewpoints.end(); ++it )
    viewpoint->removeItem(it.key());

  viewpoints.clear();
}

void MView::save_view()
{
  GLMesh_Settings s;
  int id;
  QString name;

  name.sprintf("View %d", viewpoints.count()+1);
  s = glmesh->getSettings();
  id = viewpoint->insertItem(name, this, SLOT(load_view(int)) );

  viewpoints.insert(id, s);
}

void MView::load_view(int id)
{
  glmesh->setViewpoint(viewpoints[id]);
}

void MView::setTexture()
{
  modes->setItemChecked(ids[displayMode], FALSE);
  displayMode=MVIEW_TEXTURE;
  modes->setItemChecked(ids[displayMode], TRUE);
  glmesh->setDisplayMode(TEXTURE);
}

void MView::setSolid()
{
  modes->setItemChecked(ids[displayMode], FALSE);
  displayMode=MVIEW_SOLID;
  modes->setItemChecked(ids[displayMode], TRUE);
  glmesh->setDisplayMode(SOLID);
}

void MView::setFrontlines()
{
  modes->setItemChecked(ids[displayMode], FALSE);
  displayMode=MVIEW_FRONTLINES;
  modes->setItemChecked(ids[displayMode], TRUE);
  glmesh->setDisplayMode(FRONTLINES);
}

void MView::setWire()
{
  modes->setItemChecked(ids[displayMode], FALSE);
  displayMode=MVIEW_WIRE;
  modes->setItemChecked(ids[displayMode], TRUE);
  glmesh->setDisplayMode(WIRE);
}

void MView::setPoints()
{
  modes->setItemChecked(ids[displayMode], FALSE);
  displayMode=MVIEW_POINTS;
  modes->setItemChecked(ids[displayMode], TRUE);
  glmesh->setDisplayMode(POINTS);
}

void MView::setFeatures()
{
  modes->setItemChecked(ids[displayMode], FALSE);
  displayMode=MVIEW_FEATURES;
  modes->setItemChecked(ids[displayMode], TRUE);
  glmesh->setDisplayMode(FEATURES);
}

void MView::togglePick()
{
  static CoordinatesWindow *coordinates;

  bool b_pick = !modes->isItemChecked(ids[MVIEW_PICK]);
#ifdef TEXT_MAP
  bool b_text = modes->isItemChecked(ids[MVIEW_TEXT]);
#else
  bool b_text = FALSE;
#endif

  if (b_pick)
    {
      QDesktopWidget desktop;

      coordinates = new CoordinatesWindow(this, settings->mesh,
					  "Coordinates Window");
      if ( frameGeometry().topRight().x() + coordinates->width() <= 
	   desktop.screen( desktop.primaryScreen() )->width() )

      	coordinates->move( frameGeometry().topRight().x() + 1,
			   frameGeometry().topRight().y() );

      else if ( frameGeometry().topLeft().x() - coordinates->width() >= 0 ) 

	coordinates->move( frameGeometry().topLeft().x() - 
			   coordinates->frameGeometry().width() - 8, 
			   frameGeometry().topLeft().y() );

      coordinates->show();
    }
  else
    delete coordinates;

  if (!b_pick && !b_text)
    glmesh->clearSelection();

  pick->setOn(b_pick);
  modes->setItemChecked(ids[MVIEW_PICK], b_pick);

  glmesh->setPick(b_pick || b_text, b_pick ? coordinates : NULL);
}

void MView::toggleTextureMapping()
{
#ifdef TEXT_MAP
  static TextureImageWindow *texture_image;

  bool b_pick = modes->isItemChecked(ids[MVIEW_PICK]);
  bool b_text = !modes->isItemChecked(ids[MVIEW_TEXT]);

  if (b_text)
    {
      QDesktopWidget desktop;

      glmesh->setKeepPick(1);

      texture_image = new TextureImageWindow(this, glmesh, settings->mesh,
					      "TextureImage Window");
      if ( frameGeometry().topRight().x() + texture_image->width() <= 
	   desktop.screen( desktop.primaryScreen() )->width() )

      	texture_image->move( frameGeometry().topRight().x() + 1,
			     frameGeometry().topRight().y() );

      else if ( frameGeometry().topLeft().x() - texture_image->width() > 0 ) 

	texture_image->move( frameGeometry().topLeft().x() - 
			     texture_image->frameGeometry().width() - 8, 
			     frameGeometry().topLeft().y() );
	
      texture_image->show();
    }
  else
    {
      glmesh->setKeepPick(0);
      glmesh->clearSelection();

      delete texture_image;
    }

  text->setOn(b_text);
  modes->setItemChecked(ids[MVIEW_TEXT], b_text);

  glmesh->setPick2(b_pick || b_text);
#endif
}

void MView::toggleLighting()
{
  bool b = !modes->isItemChecked(ids[MVIEW_LIGHTING]);

  modes->setItemChecked(ids[MVIEW_LIGHTING],b);
  glmesh->setLight(b);
}

void MView::toggleShapeColors()
{
  bool b = !modes->isItemChecked(ids[MVIEW_SHAPE]);

  modes->setItemChecked(ids[MVIEW_SHAPE],b);
  glmesh->setShapeColors(b);
}

void MView::toggleSurfaceNormals()
{
  bool b = !modes->isItemChecked(ids[MVIEW_NORMALS]);

  modes->setItemChecked(ids[MVIEW_NORMALS],b);
  glmesh->setNormals(b);
}

void MView::toggleCutBackFaces()
{
  bool b = !modes->isItemChecked(ids[MVIEW_CUT]);

  modes->setItemChecked(ids[MVIEW_CUT],b);
  glmesh->setBackFaces(b);
}

void MView::toggleBilinearFiltering()
{
  bool b = !modes->isItemChecked(ids[MVIEW_FILTER]);

  modes->setItemChecked(ids[MVIEW_FILTER],b);
  glmesh->setTexturefiltering(b);
}

void MView::toggleFeatures()
{
  bool b = !modes->isItemChecked(ids[MVIEW_FEATURE]);

  modes->setItemChecked(ids[MVIEW_FEATURE],b);
  glmesh->setFeatures(b);
}

void MView::toggleAspectRatio()
{
  bool b = !modes->isItemChecked(ids[MVIEW_RATIO]);

  modes->setItemChecked(ids[MVIEW_RATIO],b);
  glmesh->setAspectRatio(b);
}

void MView::setBackgroundColor()
{
  QColor c = QColorDialog::getColor(white, this, "Color dialog");

  settings->red = c.red();
  settings->green = c.green();
  settings->blue = c.blue();

  glmesh->setBackgroundColor(c.red(), c.green(), c.blue());
}

void MView::negateNormals()
{
  if (settings->mesh != NULL)
    {
      settings->mesh->negate_surface_normals();
      glmesh->updateGL();
    }
}

void MView::removeDoublePoints()
{
  if (settings->mesh != NULL)
    {
      settings->mesh->remove_double_points();

      // Map vertices and triangle centroids to mesh centroid
      settings->mesh->move_to_centre();
      settings->mesh->scale_into_normal_sphere();

      // update status and view
      set_status();
      glmesh->updateGL();
    }
}

void MView::timerEvent(QTimerEvent *e)
{
  QString s;
  
  s.sprintf(" FPS: %1.1f ", glmesh->getFps());
  label_fps->setText(s);
}

void MView::takeShot(int item)
{
  // glmesh->renderPixmap(TRUE)
  // linux software renderer: display errors afterwards
  // solaris: display errors afterwards

  // glmesh->renderPixmap()
  // linux software renderer: ok
  // solaris: error GLXBadContext and freezes mview

  // gl->renderPixmap(TRUE)
  // linux software renderer: ok
  // solaris: no error, but image saved wrong

  // gl->renderPixmap()
  // linux software renderer: ok
  // solaris: error GLXBadContext and freezes mview

  const char *format = screenshot->text(item);
  QString filename = QFileDialog::getSaveFileName(QString::null,
						  QString::null, this);
  if (!filename.isEmpty())
    {
      //GLMesh_Settings s = glmesh->getSettings();
      //GLMesh *gl = new GLMesh(NULL, "GL Mesh", s);
      //QPixmap pixmap = gl->renderPixmap(glmesh->width(),glmesh->height(),TRUE);
      //QPixmap pixmap = gl->renderPixmap(glmesh->width(),glmesh->height(),FALSE);
      //delete gl;

      QPixmap pixmap = glmesh->renderPixmap();

      QImage image = pixmap.convertToImage();

      if (image.save(filename, format) == FALSE)
	QMessageBox::warning(this, "Save screenshot", "Could not save image!");
    }
}

void MView::save_as(int item)
{
  QString name = save->text(item);
  QString filename = QFileDialog::getSaveFileName(QString::null,
						  QString::null, this);
  if (!filename.isEmpty())
    {
      Mesh *m = NULL;
      FILE *f;

      if ((f = fopen(filename, "w")) == NULL)
	return;

      if (name == "P Mesh")
	m = new P_Mesh();
      if (name == "GTS")
	m = new GTS_Mesh();
      if (name == "Geomview")
	m = new Geomview_Mesh();
      if (name == "VRML 1")
	m = new Vrml1_Mesh();
      if (name == "VRML 2")
	m = new Vrml2_Mesh();
      if (name == "PLY")
	m = new PLY_Mesh();

      if (m != NULL)
	{
	  m->set_mesh(settings->mesh);
	  m->write(f);
	  delete m;
	}

      fclose(f);
    }
}

void MView::load_mesh()
{
  QString fn = QFileDialog::getOpenFileName( QString::null, QString::null,
					     this);
  if ( !fn.isEmpty() )
    {
      delete settings->mesh;
      settings->mesh=NULL;
      glmesh->newMesh(*settings);

      name = fn;
      load_mesh_file(fn, &settings->mesh);

      if (settings->mesh != NULL)
	{
	  // Map vertices and triangle centroids to mesh centroid
	  settings->mesh->move_to_centre();
	  settings->mesh->scale_into_normal_sphere();
	}

      glmesh->newMesh(*settings);
      set_mview_defaults();
      set_status();
    }
}

void MView::load_feature()
{
  QString fn = QFileDialog::getOpenFileName( QString::null, QString::null,
                                               this);
  if ( !fn.isEmpty() )
    {
      delete settings->features;
      settings->features=NULL;
      glmesh->newFeatures(settings->features);

      load_mesh_file(fn, &settings->features);

      glmesh->newFeatures(settings->features);
      set_status();
    }
}

int MView::load_mesh_file(const char *fileName, Mesh **mesh)
{
  FILE *f;

  if ((f = fopen(fileName, "r")) == NULL)
    return 1;

  switch (Mesh::type(f))
    {
    case LIST_MESH:
      *mesh = new List_Mesh();
      break;
    case P_MESH:
      *mesh = new P_Mesh();
      break;
    case GTS_MESH:
      *mesh = new GTS_Mesh();
      break;
    case GEOMVIEW_MESH:
      *mesh = new Geomview_Mesh();
      break;
	  case PLY_MESH:
      *mesh = new PLY_Mesh();
      break;	
    case VRML1_MESH:
      *mesh = new Vrml1_Mesh();
      break;
    case VRML2_MESH:
      *mesh = new Vrml2_Mesh();
      break;
    case FEATURES_MESH:
      *mesh = new Feature_Mesh();
      break;
    default:
      fclose(f);
      return 2;
    }

  pd=new QProgressDialog("Loading...", "Cancel", 0, this, "progress", TRUE);
  pd->setCaption("Please Wait");

  if ((*mesh)->read(f, (int (*)(int)) update_progress,
		    (void (*)(int)) set_total) != 0)
    (*mesh)->clear();
  
  delete pd;
  fclose(f);

  return 0;
}

void MView::set_glmesh_defaults()
{
  settings = new GLMesh_Settings();
 
  settings->mesh = settings->features = NULL;
  settings->mesh_displayMode = SOLID;
  settings->xShift = settings->yShift = settings->zShift = 0.0;
  settings->clipping = 0.0;
  settings->point_size = 1.1;
  settings->line_size = settings->light_brightness = 1.0;
  settings->red = settings->green = settings->blue = 0.0;
  settings->f_normals = settings->f_different_colors_for_shapes = 0;
  settings->f_texture_filter = settings->f_back_faces = 0;
  settings->f_features = settings->f_keep_pick = 0;
  settings->f_lighting = settings->f_aspect_ratio = 1;
  settings->pick_x = settings->pick_y = -1;

  // create identity matrix
  for (int i=0; i < 4; i++)
    for (int j=0; j < 4; j++)
      settings->tb_transform[i][j]= i == j ? 1.0 : 0.0;
}

void MView::set_mview_defaults()
{
  displayMode=MVIEW_SOLID;

  modes->setItemChecked(ids[MVIEW_TEXTURE], FALSE);
  modes->setItemChecked(ids[MVIEW_SOLID], TRUE);
  modes->setItemChecked(ids[MVIEW_WIRE], FALSE);
  modes->setItemChecked(ids[MVIEW_FRONTLINES], FALSE);
  modes->setItemChecked(ids[MVIEW_POINTS], FALSE);
  modes->setItemChecked(ids[MVIEW_FEATURES], FALSE);
  modes->setItemChecked(ids[MVIEW_PICK], FALSE);
  modes->setItemChecked(ids[MVIEW_TEXT], FALSE);
  modes->setItemChecked(ids[MVIEW_LIGHTING], TRUE);
  modes->setItemChecked(ids[MVIEW_SHAPE], FALSE);
  modes->setItemChecked(ids[MVIEW_NORMALS], FALSE);
  modes->setItemChecked(ids[MVIEW_CUT], FALSE);
  modes->setItemChecked(ids[MVIEW_FILTER], FALSE);
  modes->setItemChecked(ids[MVIEW_FEATURE], FALSE);
  modes->setItemChecked(ids[MVIEW_RATIO], TRUE);
}

void MView::set_status()
{ 
  QString s;

  if (!name.isEmpty())
    {
      // remove directory path of the filename
      if ( name.contains('/') )
	name = name.section('/', -1);
      else if ( name.contains('\\') )
	name = name.section('\\', -1);

      s = "";

      if (settings->mesh != NULL)
	{
	  s.sprintf(": %d vertices & %d triangles ",
		    settings->mesh->number_of_vertices(),
		    settings->mesh->number_of_triangles());
	  s.insert(0, name);
	}

      label_mesh->setText(s);
  
      s = "";

      if (settings->features != NULL)
	{
	  int ver = settings->features->number_of_vertices();
	  int edg = settings->features->number_of_edges();

	  if (edg > 0)
	    s.sprintf(" Features: %d edges ", edg);
	  else
	    if (ver > 0)
	      s.sprintf(" Features: %d points ", ver);
	}

      label_features->setText(s);
    }
}

void MView::about()
{
  QMessageBox::about(this, "Mesh Viewer 0.2.2",
   "\nCopyright (c) 2001-2003 Helmut Cantzler\n\n"
   "All files in this package can be freely distributed and\n"
   "used according to the terms of the Lesser General Public License.\n\n"
   "For more information visit the MeshViewer website at:\n"
   "http://mview.sourceforge.net/\n");
}

void MView::aboutQt()
{
  QMessageBox::aboutQt(this, "Mesh Viewer");
}

